import pytest
from .merge_types import merge_types


def test_merge_types_with_primitive_types():
    schema1 = {"type": ["string", "number"]}
    schema2 = {"type": ["number", "boolean"]}
    expected = {"string", "number", "boolean"}
    assert set(merge_types(schema1, schema2)["type"]) == expected


def test_merge_types_with_same_types():
    schema1 = {"type": ["string"]}
    schema2 = {"type": ["string"]}
    expected = {"string"}
    assert set(merge_types(schema1, schema2)["type"]) == expected


def test_merge_types_with_different_single_types():
    schema1 = {"type": "string"}
    schema2 = {"type": "number"}
    expected = {"string", "number"}
    assert set(merge_types(schema1, schema2)["type"]) == expected


def test_merge_types_with_null_type():
    schema1 = {"type": "null"}
    schema2 = {"type": ["string"]}
    expected = {"string", "null"}
    assert set(merge_types(schema1, schema2)["type"]) == expected


def test_merge_types_with_non_primitive_type():
    schema1 = {"type": ["object"]}
    schema2 = {"type": ["string"]}
    expected = {"object"}  # Non-primitive type takes precedence
    assert set(merge_types(schema1, schema2)["type"]) == expected


def test_merge_types_with_both_non_primitive_types():
    schema1 = {"type": ["object"]}
    schema2 = {"type": ["array"]}
    expected = {"object"}  # Non-primitive type takes precedence
    assert set(merge_types(schema1, schema2)["type"]) == expected


def test_merge_types_with_empty_schema():
    schema1 = {"type": []}
    schema2 = {"type": ["string"]}
    expected = {"string"}
    assert set(merge_types(schema1, schema2)["type"]) == expected


def test_merge_types_with_empty_and_null_schema():
    schema1 = {"type": "null"}
    schema2 = {"type": []}
    expected = {"null"}
    assert set(merge_types(schema1, schema2)["type"]) == expected


def test_merge_types_with_empty_arrays():
    schema1 = {"type": []}
    schema2 = {"type": []}
    assert len(set(merge_types(schema1, schema2)["type"])) == 0


def test_merge_types_with_primitive_and_non_primitive():
    schema1 = {"type": ["string"]}
    schema2 = {"type": ["object"]}
    expected = {"string"}
    assert set(merge_types(schema1, schema2)["type"]) == expected

    schema1 = {"type": ["string"]}
    schema2 = {"type": ["object"]}
    expected = {"object"}
    assert set(merge_types(schema2, schema1)["type"]) == expected


def test_merge_types_with_multiple_nulls():
    schema1 = {"type": "null"}
    schema2 = {"type": "null"}
    expected = {"null"}
    assert set(merge_types(schema1, schema2)["type"]) == expected


def test_merge_types_with_primitive_array_and_null():
    schema1 = {"type": ["null"]}
    schema2 = {"type": ["string"]}
    expected = {"string", "null"}
    assert set(merge_types(schema1, schema2)["type"]) == expected



SUPPORTED_KINDS = [
    "sql_op",
    "nosql_op",
    "outgoing_http_op",
    "fs_op",
    "exec_op",
    "deserialize_op",
    "ai_op",
]


class Operations(dict):
    def __init__(self):
        super().__init__()

    def ensure_operation(self, operation, kind):
        if not kind in SUPPORTED_KINDS:
            raise Exception(f"Kind {kind} is not supported for operations.")
        if not operation in self.keys():
            self[operation] = {
                "kind": kind,
                "total": 0,
                "attacksDetected": {
                    "total": 0,
                    "blocked": 0,
                },
            }

    def register_call(self, operation, kind):
        self.ensure_operation(operation, kind)
        self[operation]["total"] += 1

    def on_detected_attack(self, blocked, operation):
        if operation not in self.keys():
            return

        self[operation]["attacksDetected"]["total"] += 1
        if blocked:
            self[operation]["attacksDetected"]["blocked"] += 1

    def update(self, m, /, **kwargs):
        for operation in m.keys():
            self.ensure_operation(operation, kind=m[operation]["kind"])

            self[operation]["total"] += m[operation]["total"]

            imported_attacks_total = m[operation]["attacksDetected"]["total"]
            self[operation]["attacksDetected"]["total"] += imported_attacks_total

            imported_attacks_blocked = m[operation]["attacksDetected"]["blocked"]
            self[operation]["attacksDetected"]["blocked"] += imported_attacks_blocked



import pytest
from .unsafe_path_start import starts_with_unsafe_path


def test_linux_root_paths():
    assert starts_with_unsafe_path("/etc/passwd", "/etc")
    assert starts_with_unsafe_path("/bin/bash", "/bin")
    assert starts_with_unsafe_path("/lib/modules", "/lib")
    assert starts_with_unsafe_path("/home/user/file.txt", "/home")
    assert starts_with_unsafe_path("/usr/local/bin", "/usr")
    assert starts_with_unsafe_path("/var/log/syslog", "/var")


def test_windows_paths():
    assert starts_with_unsafe_path("c:/Program Files/app.exe", "c:/")
    assert starts_with_unsafe_path("c:\\Windows\\System32\\cmd.exe", "c:\\")
    assert not starts_with_unsafe_path("d:/Documents/file.txt", "c:/")


def test_edge_cases():
    assert not starts_with_unsafe_path("", "/etc")
    assert not starts_with_unsafe_path("/etc", "")
    assert starts_with_unsafe_path("c:/", "c:/")
    assert starts_with_unsafe_path("c:/folder/file.txt", "c:/folder")



import pytest
from aikido_zen.helpers.try_decode_as_jwt import try_decode_as_jwt


def test_returns_false_for_empty_string():
    assert try_decode_as_jwt("") == (False, None)


def test_returns_false_for_invalid_JWT():
    assert try_decode_as_jwt("invalid") == (False, None)
    assert try_decode_as_jwt("invalid.invalid") == (False, None)
    assert try_decode_as_jwt("invalid.invalid.invalid") == (False, None)
    assert try_decode_as_jwt("invalid.invalid.invalid.invalid") == (False, None)


def test_returns_payload_for_invalid_JWT():
    assert try_decode_as_jwt("/;ping%20localhost;.e30=.") == (True, {})
    assert try_decode_as_jwt("/;ping%20localhost;.W10=.") == (True, [])


def test_returns_decoded_JWT_for_valid_JWT():
    assert try_decode_as_jwt(
        "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiIxMjM0NTY3ODkwIiwidXNlcm5hbWUiOnsiJG5lIjpudWxsfSwiaWF0IjoxNTE2MjM5MDIyfQ._jhGJw9WzB6gHKPSozTFHDo9NOHs3CNOlvJ8rWy6VrQ"
    ) == (True, {"sub": "1234567890", "username": {"$ne": None}, "iat": 1516239022})


def test_returns_decoded_JWT_for_valid_JWT_with_bearer_prefix():
    assert try_decode_as_jwt(
        "Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiIxMjM0NTY3ODkwIiwidXNlcm5hbWUiOnsiJG5lIjpudWxsfSwiaWF0IjoxNTE2MjM5MDIyfQ._jhGJw9WzB6gHKPSozTFHDo9NOHs3CNOlvJ8rWy6VrQ"
    ) == (True, {"sub": "1234567890", "username": {"$ne": None}, "iat": 1516239022})


def test_decodes_valid_jwt():
    example_payload = {"hello": "world"}
    example_jwt = "eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9.eyJoZWxsbyI6ICJ3b3JsZCJ9.tvagLDLoaiJKxOKqpBXSEGy7SYSifZhjntgm9ctpyj8"
    decoded_payload = try_decode_as_jwt(example_jwt)[1]

    assert decoded_payload == example_payload


def test_decodes_complete_valid_jwt():
    example_payload = {"hello": "world"}
    example_jwt = "eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9.eyJoZWxsbyI6ICJ3b3JsZCJ9.tvagLDLoaiJKxOKqpBXSEGy7SYSifZhjntgm9ctpyj8"
    decoded = try_decode_as_jwt(example_jwt)

    assert decoded[0] is True
    assert decoded[1] == {"hello": "world"}


def test_load_verify_valid_jwt():
    example_payload = {"hello": "world"}
    example_jwt = "eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9.eyJoZWxsbyI6ICJ3b3JsZCJ9.tvagLDLoaiJKxOKqpBXSEGy7SYSifZhjntgm9ctpyj8"

    decoded_payload = try_decode_as_jwt(example_jwt)[1]

    assert decoded_payload == example_payload


def test_decode_invalid_payload_string():
    example_jwt = "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.aGVsbG8gd29ybGQ.SIr03zM64awWRdPrAM_61QWsZchAtgDV3pphfHPPWkI"

    result = try_decode_as_jwt(example_jwt)

    assert result == (False, None)


def test_decode_with_non_mapping_payload_throws_exception():
    example_jwt = "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.MQ.Abcd"  # Invalid payload

    result = try_decode_as_jwt(example_jwt)

    assert result == (True, 1)


def test_decode_with_invalid_audience_param_throws_exception():
    example_jwt = "eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9.eyJoZWxsbyI6ICJ3b3JsZCJ9.tvagLDLoaiJKxOKqpBXSEGy7SYSifZhjntgm9ctpyj8"

    result = try_decode_as_jwt(example_jwt)

    assert result == (True, {"hello": "world"})


def test_decode_with_nonlist_aud_claim_throws_exception():
    example_jwt = "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJoZWxsbyI6IndvcmxkIiwiYXVkIjoxfQ.Rof08LBSwbm8Z_bhA2N3DFY-utZR1Gi9rbIS5Zthnnc"

    result = try_decode_as_jwt(example_jwt)

    assert result == (True, {"hello": "world", "aud": 1})


def test_decode_with_invalid_aud_list_member_throws_exception():
    example_jwt = (
        "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9"
        ".eyJoZWxsbyI6IndvcmxkIiwiYXVkIjpbMV19"
        ".iQgKpJ8shetwNMIosNXWBPFB057c2BHs-8t1d2CCM2A"
    )

    result = try_decode_as_jwt(example_jwt)

    assert result == (
        True,
        {"hello": "world", "aud": [1]},
    )  # Assuming the function handles audience checks internally


def test_decode_raises_exception_if_exp_is_not_int():
    example_jwt = (
        "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9."
        "eyJleHAiOiJub3QtYW4taW50In0."
        "P65iYgoHtBqB07PMtBSuKNUEIPPPfmjfJG217cEE66s"
    )

    result = try_decode_as_jwt(example_jwt)

    assert result == (
        True,
        {"exp": "not-an-int"},
    )  # Assuming the function handles exp checks internally


def test_decode_raises_exception_if_iat_is_not_int():
    example_jwt = (
        "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9."
        "eyJpYXQiOiJub3QtYW4taW50In0."
        "H1GmcQgSySa5LOKYbzGm--b1OmRbHFkyk8pq811FzZM"
    )

    result = try_decode_as_jwt(example_jwt)

    assert result == (
        True,
        {"iat": "not-an-int"},
    )  # Assuming the function handles iat checks internally


def test_decodes_valid_es256_jwt():
    example_payload = {"hello": "world"}
    example_jwt = (
        "eyJhbGciOiJFUzI1NiIsInR5cCI6IkpXVCJ9."
        "eyJoZWxsbyI6IndvcmxkIn0."
        "TORyNQab_MoXM7DvNKaTwbrJr4UYd2SsX8hhlnWelQFmPFSf_JzC2EbLnar92t-bXsDovzxp25ExazrVHkfPkQ"
    )

    result = try_decode_as_jwt(example_jwt)

    assert result == (
        True,
        example_payload,
    )  # Assuming the function handles decoding correctly


def test_decodes_valid_rs384_jwt():
    example_payload = {"hello": "world"}
    example_jwt = (
        "eyJhbGciOiJSUzM4NCIsInR5cCI6IkpXVCJ9."
        "eyJoZWxsbyI6IndvcmxkIn0."
        "yNQ3nI9vEDs7lEh-Cp81McPuiQ4ZRv6FL4evTYYAh1XlRTTR3Cz8pPA9Stgso8Ra9xGB4X3rlra1c8Jz10nTUjuO06OMm7oXdrnxp1KIiAJDerWHkQ7l3dlizIk1bmMA457W2fNzNfHViuED5ISM081dgf_a71qBwJ_yShMMrSOfxDx"
        "mX9c4DjRogRJG8SM5PvpLqI_Cm9iQPGMvmYK7gzcq2cJurHRJDJHTqIdpLWXkY7zVikeen6FhuGyn060Dz9gYq9tuwmrtSWCBUjiN8sqJ00CDgycxKqHfUndZbEAOjcCAhBrqWW3mSVivUfubsYbwUdUG3fSRPjaUPcpe8A"
    )

    result = try_decode_as_jwt(example_jwt)

    assert result == (
        True,
        example_payload,
    )  # Assuming the function handles decoding correctly


def test_with_urlsafe_parts():
    example_payload = {"test": "??????", "username": {"$regex": ".*"}}
    example_jwt = "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJ0ZXN0IjoiPz8_Pz8_IiwidXNlcm5hbWUiOnsiJHJlZ2V4IjoiLioifX0.DTGfM7dmKjAByZzeugmtUSpV1v9RbIsDCld2M9DlMvk"
    result = try_decode_as_jwt(example_jwt)

    assert result == (
        True,
        example_payload,
    )  # Assuming the function handles decoding correctly



"""Exports update_route_info function"""

from aikido_zen.helpers.logging import logger
from .get_api_info import get_api_info
from .merge_data_schemas import merge_data_schemas
from .merge_auth_types import merge_auth_types

ANALYSIS_ON_FIRST_X_ROUTES = 20


def update_route_info_from_context(context, route):
    if context is None:
        return
    if route["hits"] > ANALYSIS_ON_FIRST_X_ROUTES:
        return
    new_apispec = get_api_info(context)
    route["apispec"] = update_api_info(new_apispec, route["apispec"])


def update_route_info(new_apispec, route):
    """
    Checks if a route still needs to be updated (only analyzes first x routes),
    and if so, updates route using update_api_info
    """
    if route["hits"] <= ANALYSIS_ON_FIRST_X_ROUTES:
        # Only analyze the first x routes for api discovery
        route["apispec"] = update_api_info(new_apispec, route["apispec"])


def update_api_info(new_apispec, existing_apispec=None):
    """
    Merges two apispec objects into one, getting all properties from both schemas to capture optional properties.
    If the body info is not defined, the existing body info is returned (if any).
    If there is no existing body info, but the new body info is defined, the new body info is returned without merging.
    """
    try:
        if not new_apispec:
            return existing_apispec
        if not existing_apispec:
            return new_apispec

        # Body :
        existing_body = existing_apispec.get("body")
        if existing_body and new_apispec.get("body"):
            new_schema = new_apispec["body"].get("schema")
            existing_apispec["body"] = {
                "type": new_apispec["body"]["type"],
                "schema": merge_data_schemas(existing_body["schema"], new_schema),
            }
        elif new_apispec.get("body"):
            existing_apispec["body"] = new_apispec["body"]

        # Query :
        existing_query = existing_apispec.get("query")
        if existing_query and new_apispec.get("query"):
            existing_apispec["query"] = merge_data_schemas(
                existing_query, new_apispec["query"]
            )
        elif new_apispec.get("query"):
            existing_apispec["query"] = new_apispec["query"]

        # Auth :
        existing_apispec["auth"] = merge_auth_types(
            existing_apispec.get("auth"), new_apispec.get("auth")
        )
        return existing_apispec
    except Exception as e:
        logger.debug("Exception occurred in update_api_info : %s", e)



"""Mainly exports update_service_config function"""

from aikido_zen.helpers.logging import logger
from aikido_zen.helpers.get_current_unixtime_ms import get_unixtime_ms


def update_service_config(connection_manager, res):
    """
    Update configuration based on the server's response
    """
    if res.get("success", False) is False:
        logger.debug(res)
        return
    if "block" in res.keys() and res["block"] != connection_manager.block:
        logger.debug("Updating blocking, setting blocking to : %s", res["block"])
        connection_manager.block = bool(res["block"])

    connection_manager.conf.update(
        endpoints=res.get("endpoints", []),
        last_updated_at=res.get("configUpdatedAt", get_unixtime_ms()),
        blocked_uids=res.get("blockedUserIds", []),
        bypassed_ips=res.get("allowedIPAddresses", []),
        received_any_stats=res.get("receivedAnyStats", True),
    )



import pytest
from .serialize_to_json import serialize_to_json


def test_serialize_basic_types():
    assert serialize_to_json({"key": "value"}) == '{"key": "value"}'
    assert serialize_to_json(["item1", "item2"]) == '["item1", "item2"]'
    assert serialize_to_json(123) == "123"
    assert serialize_to_json(45.67) == "45.67"
    assert serialize_to_json(True) == "true"
    assert serialize_to_json(None) == "null"


def test_serialize_complex_types():
    class CustomObject:
        def __init__(self, name):
            self.name = name

    obj = CustomObject("test")
    assert serialize_to_json(obj) == "{}"


def test_serialize_dict_with_non_serializable_key():
    obj = {object(): "value"}
    assert serialize_to_json(obj) == "{}"


def test_serialize_list_with_non_serializable_item():
    obj = [1, 2, object()]
    assert serialize_to_json(obj) == "{}"


def test_serialize_nested_structures():
    obj = {"key": ["value1", {"nested_key": "nested_value"}]}
    assert (
        serialize_to_json(obj) == '{"key": ["value1", {"nested_key": "nested_value"}]}'
    )


def test_serialize_empty_input():
    assert serialize_to_json("") == '""'
    assert serialize_to_json([]) == "[]"
    assert serialize_to_json({}) == "{}"


def test_serialize_non_serializable_object():
    class NonSerializable:
        pass

    obj = NonSerializable()
    assert serialize_to_json(obj) == "{}"



"""Helper function file, see docstring"""

MOVED_PERMANENTLY = 301
FOUND = 302
SEE_OTHER = 303
TEMPORARY_REDIRECT = 307
PERMANENT_REDIRECT = 308

REDIRECT_CODES = [
    MOVED_PERMANENTLY,
    FOUND,
    SEE_OTHER,
    TEMPORARY_REDIRECT,
    PERMANENT_REDIRECT,
]


def is_redirect_status_code(status_code):
    """Checks if this status code is a redirect code"""
    return status_code in REDIRECT_CODES



"""Exports function `get_clean_stacktrace`"""

import inspect
import sys


def get_clean_stacktrace():
    """Returns a cleaned up stacktrace"""
    # Get the current stack
    stack = inspect.stack()

    # List of built-in modules to filter out
    ignored_modules = sys.builtin_module_names

    cleaned_stack = []

    for frame_info in stack:
        name = frame_info.frame.f_globals.get("__name__", "")

        if name not in ignored_modules and not name.startswith("aikido_zen"):
            cleaned_stack.append(
                f"File: {frame_info.filename}, L{frame_info.lineno} {frame_info.function}(...)"
            )

    cleaned_stack.reverse()
    return "• " + "\n\n• ".join(cleaned_stack)



"""Helper function file, exports normalize_url"""

from urllib.parse import urlparse, urlunparse


def normalize_url(url):
    """Normalizes the url"""
    # Parse the URL
    parsed_url = urlparse(url)

    # Normalize components
    scheme = parsed_url.scheme.lower()  # Lowercase scheme
    netloc = parsed_url.netloc.lower()  # Lowercase netloc
    path = parsed_url.path.rstrip("/")  # Remove trailing slash
    query = parsed_url.query  # Keep query as is
    fragment = parsed_url.fragment  # Keep fragment as is

    # Remove default ports (80 for http, 443 for https)
    if scheme == "http" and parsed_url.port == 80:
        netloc = netloc.replace(":80", "")
    elif scheme == "https" and parsed_url.port == 443:
        netloc = netloc.replace(":443", "")

    # Reconstruct the normalized URL
    normalized_url = urlunparse((scheme, netloc, path, "", query, fragment))
    return normalized_url



"""
Sink module for `psycopg2`
"""

from aikido_zen.background_process.packages import is_package_compatible
import aikido_zen.vulnerabilities as vulns
from aikido_zen.helpers.get_argument import get_argument
from aikido_zen.helpers.register_call import register_call
from aikido_zen.sinks import on_import, before, patch_function, after


@after
def _connect(func, instance, _args, _kwargs, rv):
    """
    the return value (rv) is the new connection object, we'll change the cursor_factory attribute here.
    """
    if rv.cursor_factory is None:
        # set default if not set
        import psycopg2.extensions

        rv.cursor_factory = psycopg2.extensions.cursor

    rv.cursor_factory = type(
        "AikidoPsycopg2Cursor",
        (rv.cursor_factory,),
        {
            # Allows us to modify these otherwise immutable functions
            "execute": rv.cursor_factory.execute,
            "executemany": rv.cursor_factory.executemany,
        },
    )
    patch_function(rv.cursor_factory, "execute", psycopg2_patch)
    patch_function(rv.cursor_factory, "executemany", psycopg2_patch)


@before
def psycopg2_patch(func, instance, args, kwargs):
    query = get_argument(args, kwargs, 0, "query")

    op = f"psycopg2.Connection.Cursor.{func.__name__}"
    register_call(op, "sql_op")

    vulns.run_vulnerability_scan(kind="sql_injection", op=op, args=(query, "postgres"))


@on_import("psycopg2")
def patch(m):
    """
    patching module psycopg2
    - patches psycopg2.connect
    cannot set 'execute' attribute of immutable type 'psycopg2.extensions.cursor',
    so we create our own cursor factory to bypass this limitation.
    """
    compatible = is_package_compatible(
        required_version="2.9.2", packages=["psycopg2", "psycopg2-binary"]
    )
    if not compatible:
        # Users can install either psycopg2 or psycopg2-binary, we need to check if at least
        # one is installed and if they meet version requirements
        return

    patch_function(m, "connect", _connect)



"""Exports get_argument"""


def get_argument(args, kwargs, pos, name):
    """Checks kwargs and args for your argument"""
    if name in kwargs:
        return kwargs.get(name)
    if args and len(args) > pos:
        return args[pos]
    return None



"""Helper function file, see function docstring"""

import os
import hashlib


def hash_aikido_token():
    """Creates a hash of AIKIDO_TOKEN"""
    token = os.getenv("AIKIDO_TOKEN")

    if isinstance(token, str):
        hash_object = hashlib.sha256(token.encode("utf-8"))
        return hash_object.hexdigest()
    hash_object = hashlib.sha256("default".encode("utf-8"))
    return hash_object.hexdigest()



"""Exports get_api_info function"""

from aikido_zen.helpers.logging import logger
from aikido_zen.helpers.env_vars.feature_flags import is_feature_enabled
from .get_body_data_type import get_body_data_type
from .get_data_schema import get_data_schema
from .get_auth_types import get_auth_types


def get_api_info(context):
    """Generates an apispec based on the context passed along"""
    try:
        body_info = get_body_info(context)
        query_info = get_query_info(context)
        auth_info = get_auth_types(context)
        if body_info or query_info or auth_info:
            return {"body": body_info, "query": query_info, "auth": auth_info}
    except Exception as e:
        logger.debug("Exception occurred whilst generating apispec: %s", e)

    return {}


def get_body_info(context):
    """Returns type, schema dict with body info for the given context"""
    data = context.body
    if not data or not isinstance(data, dict):
        data = context.xml
    if not data or not isinstance(data, dict):
        return None
    return {
        "type": get_body_data_type(context.headers),
        "schema": get_data_schema(data),
    }


def get_query_info(context):
    """If context has query data this returns it's schema"""
    query_info = None
    if context.query and isinstance(context.query, dict) and len(context.query) > 0:
        query_info = get_data_schema(context.query)
    return query_info



import pytest
from .merge_auth_types import merge_auth_types as merge


def test_merge_api_auth_types():
    result = merge(
        [
            {
                "type": "http",
                "scheme": "bearer",
            },
            {
                "type": "apiKey",
                "in": "header",
                "name": "x-api-key",
            },
        ],
        [
            {
                "type": "http",
                "scheme": "bearer",
            },
            {
                "type": "http",
                "scheme": "basic",
            },
            {
                "type": "apiKey",
                "in": "header",
                "name": "x-api-key-v2",
            },
        ],
    )
    assert result == [
        {
            "type": "http",
            "scheme": "bearer",
        },
        {
            "type": "apiKey",
            "in": "header",
            "name": "x-api-key",
        },
        {
            "type": "http",
            "scheme": "basic",
        },
        {
            "type": "apiKey",
            "in": "header",
            "name": "x-api-key-v2",
        },
    ]

    assert merge(None, None) is None

    assert merge(
        [
            {
                "type": "http",
                "scheme": "bearer",
            },
        ],
        None,
    ) == [
        {
            "type": "http",
            "scheme": "bearer",
        },
    ]

    assert merge(
        None,
        [
            {
                "type": "http",
                "scheme": "digest",
            },
        ],
    ) == [
        {
            "type": "http",
            "scheme": "digest",
        },
    ]


# To run the tests, use the command: pytest <filename>.py



"""
This will check the context of the request for SQL Injections
"""

from aikido_zen.helpers.extract_strings_from_context import extract_strings_from_context
from aikido_zen.vulnerabilities.sql_injection import detect_sql_injection


def context_contains_sql_injection(sql, operation, context, dialect):
    """
    This will check the context of the request for SQL Injections
    """
    if not isinstance(sql, str):
        # Only supports SQL queries that are strings, return otherwise.
        return {}
    for user_input, path, source in extract_strings_from_context(context):
        if detect_sql_injection(sql, user_input, dialect):
            return {
                "operation": operation,
                "kind": "sql_injection",
                "source": source,
                "pathToPayload": path,
                "metadata": {
                    "sql": sql,
                    "dialect": dialect,
                },
                "payload": user_input,
            }
    return {}



import pytest
from .aikido_disabled_flag_active import aikido_disabled_flag_active


def test_aikido_disabled_flag_does_not_exist(monkeypatch):
    monkeypatch.setenv("AIKIDO_DISABLE", None)
    assert aikido_disabled_flag_active() == False


def test_aikido_disabled_flag_exists_but_invalid(monkeypatch):
    monkeypatch.setenv("AIKIDO_DISABLE", "token")
    assert aikido_disabled_flag_active() == False


def test_aikido_disabled_flag_exists_and_off(monkeypatch):
    monkeypatch.setenv("AIKIDO_DISABLE", "0")
    assert aikido_disabled_flag_active() == False

    monkeypatch.setenv("AIKIDO_DISABLE", "false")
    assert aikido_disabled_flag_active() == False

    monkeypatch.setenv("AIKIDO_DISABLE", "False")
    assert aikido_disabled_flag_active() == False


def test_aikido_disabled_flag_exists_and_on(monkeypatch):
    monkeypatch.setenv("AIKIDO_DISABLE", "1")
    assert aikido_disabled_flag_active() == True

    monkeypatch.setenv("AIKIDO_DISABLE", "True")
    assert aikido_disabled_flag_active() == True

    monkeypatch.setenv("AIKIDO_DISABLE", "TRUE")
    assert aikido_disabled_flag_active() == True

    monkeypatch.setenv("AIKIDO_DISABLE", "true")
    assert aikido_disabled_flag_active() == True



"""
Simply exports the aikido background process
"""

import multiprocessing.connection as con
import time
import sched
import traceback
import sys
from threading import Thread
from queue import Queue
from aikido_zen.helpers.logging import logger
from aikido_zen.background_process.cloud_connection_manager import (
    CloudConnectionManager,
)
from aikido_zen.helpers.check_env_for_blocking import check_env_for_blocking
from aikido_zen.helpers.token import get_token_from_env
from aikido_zen.background_process.api.http_api_ratelimited import (
    ReportingApiHTTPRatelimited,
)
from aikido_zen.helpers.urls.get_api_url import get_api_url
from .commands import process_incoming_command

EMPTY_QUEUE_INTERVAL = 5  # 5 seconds


class AikidoBackgroundProcess:
    """
    Aikido's background process consists of 2 threads :
    - (main) Listening thread which listens on an IPC socket for incoming data
    - (spawned) reporting thread which collects IPC data and send it to a CloudConnectionManager
    """

    def __init__(self, address, key):
        try:
            listener = con.Listener(address, authkey=None)
        except OSError:
            logger.debug("not listening, another thread is already listening.")
            sys.exit(0)
        self.queue = Queue()
        self.connection_manager = None
        # Start reporting thread :
        Thread(target=self.reporting_thread).start()

        logger.debug(
            "Background process started successfully, with UDS File : %s", address
        )

        while True:
            conn = listener.accept()
            while True:
                try:
                    data = conn.recv()  #  because of this no sleep needed in thread
                    logger.debug("Incoming data : %s", data)
                    process_incoming_command(
                        connection_manager=self.connection_manager,
                        obj=data,
                        conn=conn,
                        queue=self.queue,
                    )
                    conn.close()  # Sort of EOL for Python IPC
                    break
                except Exception as e:
                    logger.error("Exception occurred in server thread : %s", e)
                    logger.debug("Trace \n %s", traceback.format_exc())
                    break  # Return back to listening for new connections

    def reporting_thread(self):
        """Reporting thread"""
        logger.debug("Started reporting thread")
        event_scheduler = sched.scheduler(
            time.monotonic, time.sleep
        )  # Create an event scheduler
        self.send_to_connection_manager(event_scheduler)

        api = ReportingApiHTTPRatelimited(
            reporting_url=get_api_url(),
            max_events_per_interval=100,
            interval_in_ms=60 * 60 * 1000,
        )
        # We need to pass along the scheduler so that the heartbeat also gets sent
        self.connection_manager = CloudConnectionManager(
            block=check_env_for_blocking(),
            api=api,
            token=get_token_from_env(),
            serverless=False,
        )
        time.sleep(2)  # Sleep 2 seconds to make sure modules get reported
        self.connection_manager.start(event_scheduler)
        event_scheduler.run()

    def send_to_connection_manager(self, event_scheduler):
        """
        Reports the found data to an Aikido server
        """
        # Add back to event scheduler in EMPTY_QUEUE_INTERVAL secs :
        event_scheduler.enter(
            EMPTY_QUEUE_INTERVAL, 1, self.send_to_connection_manager, (event_scheduler,)
        )
        while not self.queue.empty():
            queue_attack_item = self.queue.get()
            self.connection_manager.on_detected_attack(
                attack=queue_attack_item[0],
                context=queue_attack_item[1],
                blocked=queue_attack_item[2],
                stack=queue_attack_item[3],
            )



import pytest
from aikido_zen.context.asgi import set_asgi_attributes_on_context


class Context:
    def __init__(self):
        self.headers = None
        self.cookies = None
        self.method = None
        self.url = None
        self.query = None
        self.remote_address = None


# Scope 1 :
TEST_ASGI_SCOPE_1 = {
    "method": "PUT",
    "headers": [(b"COOKIE", b"a=b; c=d"), (b"header1_test-2", b"testValue2198&")],
    "query_string": b"a=b&b=d",
    "client": ["1.1.1.1"],
    "server": ["192.168.0.1", 443],
    "scheme": "https",
    "root_path": "192.168.0.1",
    "path": "192.168.0.1/a/b/c/d",
}


def test_asgi_scope_1():
    context1 = Context()
    set_asgi_attributes_on_context(context1, TEST_ASGI_SCOPE_1)
    assert context1.method == "PUT"
    assert context1.remote_address == "1.1.1.1"
    assert context1.query == {"a": ["b"], "b": ["d"]}
    assert context1.headers == {
        "COOKIE": "a=b; c=d",
        "HEADER1_TEST_2": "testValue2198&",
    }
    assert context1.cookies == {"a": "b", "c": "d"}
    assert context1.url == "https://192.168.0.1:443/a/b/c/d"


# Scope 2 :
TEST_ASGI_SCOPE_2 = {
    "method": "GET",
    "headers": [(b"COOKIE", b"x=y; z=w"), (b"header2_test-1", b"anotherValue")],
    "query_string": b"x=y&z=w",
    "client": ["2.2.2.2"],
    "server": ["192.168.0.2", 80],
    "scheme": "http",
    "root_path": "",
    "path": "/path/to/resource",
}


def test_asgi_scope_2():
    context2 = Context()
    set_asgi_attributes_on_context(context2, TEST_ASGI_SCOPE_2)
    assert context2.method == "GET"
    assert context2.remote_address == "2.2.2.2"
    assert context2.query == {"x": ["y"], "z": ["w"]}
    assert context2.headers == {
        "COOKIE": "x=y; z=w",
        "HEADER2_TEST_1": "anotherValue",
    }
    assert context2.cookies == {"x": "y", "z": "w"}
    assert context2.url == "http://192.168.0.2:80/path/to/resource"


# Scope 3 :
TEST_ASGI_SCOPE_3 = {
    "method": "POST",
    "headers": [(b"COOKIE", b"session=abc123"), (b"header3_test-3", b"postValue")],
    "query_string": b"key1=value1&key2=value2",
    "client": ["3.3.3.3"],
    "server": ["192.168.0.3", 8080],
    "scheme": "http",
    "root_path": "/api",
    "path": "/api/v1/resource",
}


def test_asgi_scope_3():
    context3 = Context()
    set_asgi_attributes_on_context(context3, TEST_ASGI_SCOPE_3)
    assert context3.method == "POST"
    assert context3.remote_address == "3.3.3.3"
    assert context3.query == {"key1": ["value1"], "key2": ["value2"]}
    assert context3.headers == {
        "COOKIE": "session=abc123",
        "HEADER3_TEST_3": "postValue",
    }
    assert context3.cookies == {"session": "abc123"}
    assert context3.url == "http://192.168.0.3:8080/v1/resource"


# Scope 4 :
TEST_ASGI_SCOPE_4 = {
    "method": "DELETE",
    "headers": [(b"header4_test-4", b"deleteValue")],
    "query_string": b"",
    "client": ["4.4.4.4"],
    "server": ["192.168.0.4", 443],
    "scheme": "https",
    "root_path": "/secure",
    "path": "/secure/resource/123",
}


def test_asgi_scope_4():
    context4 = Context()
    set_asgi_attributes_on_context(context4, TEST_ASGI_SCOPE_4)
    assert context4.method == "DELETE"
    assert context4.remote_address == "4.4.4.4"
    assert context4.query == {}
    assert context4.headers == {
        "HEADER4_TEST_4": "deleteValue",
    }
    assert context4.cookies == {}  # No cookies in this scope
    assert context4.url == "https://192.168.0.4:443/resource/123"
